unit Flic;

interface

{$I KOLDEF.INC}

uses Windows, KOL, Animation;

type
  PFlic = ^TFlic;
  TFlic = object( TAnimation )
  private
    function GetPostage: PBitmap;
  protected
    FFrames: PList;
    FPostage: PBitmap;
    fSpeed: Integer;
    function GetFrameCount: Integer; virtual;
    function GetDelays(Idx: Integer): Integer; virtual;
  public
    LoadPostageStop: Boolean;
    procedure Clear; virtual;
    function LoadFromStream( Stream : PStream ) : Boolean; virtual;
    procedure StretchDraw( DC: HDC; const Dest: TRect ); virtual;
    procedure StretchDrawTransp( DC: HDC; const Dest: TRect ); virtual;
    property Postage: PBitmap read GetPostage;
  end;

function NewFlic: PFlic;

implementation

function NewFlic: PFlic;
begin
  new( Result, Create );
end;

type TFlicHeader = packed record
       size       : DWORD; // entire size of flic including this header
       typ_       : WORD;  // AF11, AF12, AF44, AF30, ????
       frames     : WORD;  // N frames in 1st segment
       width      : WORD;
       height     : WORD;
       depth      : WORD;  // bits per pixel (8 usually)
       flags      : WORD;  // 0 or 3
       speed      : DWORD; // delay between frames
       reserved1  : WORD;  // 0
       created    : DWORD;
       creator    : DWORD; // 'EGI'#0 - for EGI
       updated    : DWORD;
       updater    : DWORD;
       aspect_dx  : WORD;
       aspect_dy  : WORD;
       ext_flags  : WORD; // EGI:
                          // 1 - segment table present
                          // 2 - key images present
                          // 4 - palette is the same for all images
                          // 8 - Huffman compression
                          // 16 - BTW compression
                          // 32 - bitmap masks present
                          // 64 - multilevel masks present
                          // 128 - password protected
                          // 256 - digitized audio present
       keyframes  : WORD;
       totalframes: WORD; // EGI only
       req_mem    : DWORD;
       reserved2  : array[ 0..27 ] of Byte;
       oframe1    : DWORD; // FLC: offset to frame 1
       oframe2    : DWORD; // FLC: offset to frame 2
       reserved3  : array[ 0..39 ] of Byte;
     end;

  TChunk = packed record
    sz: DWORD;
    hd: WORD;
  end;

  TPostage = packed record
    h_Postage, w_Postage: Word;
    xlate: Word;
  end;

{ TFlic }

procedure TFlic.Clear;
begin
  FFrames.ReleaseObjects;
  FFrames := nil;
  Free_And_Nil( FPostage );
end;

function TFlic.GetDelays(Idx: Integer): Integer;
var Frm: PBitmap;
begin
  Result := fSpeed;
  if (Idx >= 0) and (Idx < Count) then
  begin
    Frm := FFrames.Items[ Idx ];
    if Frm.Tag <> 0 then
      Result := Frm.Tag;
  end;
end;

function TFlic.GetFrameCount: Integer;
begin
  Result := 0;
  if FFrames <> nil then Result := FFrames.Count;
end;

function TFlic.GetPostage: PBitmap;
begin
  Result := FPostage;
  if (Result = nil) and (Count > 0) then
    Result := FFrames.Items[ 0 ];
end;

function TFlic.LoadFromStream(Stream: PStream): Boolean;
var FliHdr: TFlicHeader;
    StartPos, NxPos, Nx1: DWORD;
    Chunk: TChunk;
    Bmp: PBitmap;

    procedure FLI_COLOR( color_shift: Integer );
    var Npackets, NSkip, NReplace: Word;
        Idx, i: Integer;
        RGB: TRGBQuad;
    begin
      Stream.Read( Npackets, 2 );
      Idx := 0;
      while Npackets > 0 do
      begin
        NSkip := 0; Stream.Read( NSkip, 1 );
        NReplace := 0; Stream.Read( NReplace, 1 );
        if NReplace = 0 then NReplace := 256;
        Inc( Idx, NSkip );
        for i := 1 to NReplace do
        begin
          RGB.rgbReserved := 0;
          //Stream.Read( RGB, 3 );
          Stream.Read( RGB.rgbRed, 1 );
          Stream.Read( RGB.rgbGreen, 1 );
          Stream.Read( RGB.rgbBlue, 1 );
          if color_shift > 0 then
          begin
            RGB.rgbBlue := RGB.rgbBlue shl color_shift;
            RGB.rgbGreen := RGB.rgbGreen shl color_shift;
            RGB.rgbRed := RGB.rgbRed shl color_shift;
          end;
          if Idx < 256 then
            Bmp.DIBPalEntries[ Idx ] := TColor( RGB );
          Inc( Idx );
        end;
        Dec( Npackets );
      end;
    end;

    procedure FLI_BLACK;
    var y: Integer;
        Dst: PByte;
    begin
      for y := 0 to Bmp.Height-1 do
      begin
        Dst := Bmp.ScanLine[ y ];
        FillChar( Dst^, Bmp.Width, 0 );
      end;
    end;

    procedure FLI_BRUN;
    var y, x: Integer;
        Dst: PByte;
        b: Integer;
        c: Byte;
        Buffer: array[ 0..255 ] of Byte;
    begin
      y := 0;
      while (y < Bmp.Height) and (Stream.Position < Nx1) do
      begin
        x := 0;
        Stream.Position := Stream.Position + 1;
        while (x < Bmp.Width) and (Stream.Position < Nx1) do
        begin
          b := 0; Stream.Read( b, 1 );
          if b and $80 <> 0 then b := b or Integer( $FFFFFF00 );
          Dst := Pointer( Integer( Bmp.ScanLine[ y ] ) + x );
          if b < 0 then
          begin
            b := -b;
            Stream.Read( Buffer, b );
            if x + b <= Bmp.Width then
              Move( Buffer, Dst^, b );
          end
            else
          if b > 0 then
          begin
            Stream.Read( c, 1 );
            if x + b <= Bmp.Width then
              FillChar( Dst^, b, c );
          end;
          Inc( x, b );
        end;
        Inc( y );
      end;
    end;

    procedure FLI_COPY;
    var y: Integer;
    Dst: PByte;
    begin
      for y := 0 to Bmp.Height-1 do
      begin
        Dst := Bmp.ScanLine[ y ];
        Stream.Read( Dst^, Bmp.Width );
        if Stream.Position >= Nx1 then break;
      end;
    end;

    procedure FLI_LC;
    var y1, h1: Word;
        y: Integer;
        n, c: Byte;
        x: Integer;
        b: Integer;
        Dst: PByte;
        Buffer: array[ 0..255 ] of Byte;
    begin
      Stream.Read( y1, 2 );
      Stream.Read( h1, 2 );
      for y := y1 to y1 + h1 - 1 do
      begin
        Stream.Read( n, 1 );
        x := 0;
        while (n > 0) and (Stream.Position < Nx1) do
        begin
          b := 0; Stream.Read( b, 1 );
          x := x + b;
          Stream.Read( b, 1 );
          if b and $80 <> 0 then b := b or Integer( $FFFFFF00 );
          Dst := Pointer( Integer( Bmp.ScanLine[ y ] ) + x );
          if b < 0 then
          begin
            b := -b;
            Stream.Read( c, 1 );
            if x + b <= Bmp.Width then
              FillChar( Dst^, b, c );
          end
            else
          begin
            Stream.Read( Buffer, b );
            if x + b <= Bmp.Width then
              Move( Buffer, Dst^, b );
          end;
          Inc( x, b );
          Dec( n );
        end;
        if Stream.Position >= Nx1 then break;
      end;
    end;

    procedure FLI_SS2;
    var nlines, w, Npackets: Word;
        y: Integer;
        n: Byte;
        c: Word;
        x: Integer;
        b: Integer;
        Dst: PByte;
        Buffer: array[ 0..255 ] of Word;
    begin
      Stream.Read( nlines, 2 );
      y := 0;
      while (nlines > 0) and (Stream.Position < Nx1) do
      begin
        Npackets := 0; // silency for compiler
        REPEAT
          Stream.Read( w, 2 );
          CASE w shr 14 OF
          0: Npackets := w;
          2: if y < Bmp.Height-1 then
             begin
               Dst := Pointer( Integer( Bmp.ScanLine[ y ] ) + Bmp.Width - 1 );
               Dst^ := w;
             end;
          3: begin
               b := -Smallint( w );
               Inc( y, b );
             end;
          END;
        UNTIL w and $C000 = 0;
        x := 0;
        while Npackets > 0 do
        begin
          Stream.Read( n, 1 );
          Inc( x, n );
          b := 0; Stream.Read( b, 1 );
          if b and $80 <> 0 then b := b or Integer( $FFFFFF00 );
          Dst := Pointer( Integer( Bmp.ScanLine[ y ] ) + x );
          if b > 0 then
          begin
            Stream.Read( Buffer, b * 2 );
            if x + b * 2 <= Bmp.Width then
              Move( Buffer, Dst^, b * 2 );
          end
            else
          if b < 0 then
          begin
            Stream.Read( c, 2 );
            if x + b * 2 <= Bmp.Width then
            while b > 0 do
            begin
              PWord( Dst )^ := c;
              Inc( Dst, 2 );
              Dec( b );
            end;
          end;
          Inc( x, b * 2 );
          Dec( Npackets );
        end;
        Dec( nlines );
      end;
    end;

    procedure FLI_PSTAMP;
    var PostHdr: TPostage;
        red, green, blue, palentry: Integer;
        RGB: TRGBQuad;
        SaveBmp: PBitmap;
        SaveNx1: DWORD;
    begin
      Stream.Read( PostHdr, Sizeof( PostHdr ) );
      Free_And_Nil( FPostage ); //  2  postage-bitmap' ...
      SaveBmp := Bmp;
      SaveNx1 := Nx1;
      TRY
        Stream.Read( Chunk, Sizeof( Chunk ) );
        if (Chunk.hd <> 15) and (Chunk.hd <> 16) then Exit;
        Nx1 := Min( Nx1, Stream.Position - Sizeof( Chunk ) + Chunk.sz );
        if Nx1 < Stream.Position then Nx1 := Stream.Position;
        FPostage := NewDibBitmap( PostHdr.w_Postage, PostHdr.h_Postage, pf8bit );
        Bmp := FPostage;
        palentry := 0;
        for red := 0 to 5 do
        for green := 0 to 5 do
        for blue := 0 to 5 do
        begin
          RGB.rgbRed := red * 256 div 6;
          RGB.rgbGreen := green * 256 div 6;
          RGB.rgbBlue := blue * 256 div 6;
          FPostage.DIBPalEntries[ palentry ] := TColor( RGB );
          Inc( palentry );
        end;
        CASE Chunk.hd OF
        15: FLI_BRUN;
        16: FLI_COPY;
        //18: FPS_XLAT256;
        END;
      FINALLY
        Bmp := SaveBmp;
        Nx1 := SaveNx1;
      END;
    end;

var Frm: PBitmap;
    FrameDelay: WORD;
    nsubchunks: Word;
begin
  Clear;
  StartPos := Stream.Position;
  Result := FALSE;
  TRY
    if Stream.Read( FliHdr, Sizeof( FliHdr ) ) <> Sizeof( FliHdr ) then Exit;
    if (FliHdr.typ_ <> $AF11) and (FliHdr.typ_ <> $AF12) and
       (FliHdr.typ_ <> $AF44) and (FliHdr.typ_ <> $AF30) then Exit;
    if FliHdr.depth > 8 then Exit;
    FWidth := FliHdr.width; if FWidth = 0 then FWidth := 320;
    FHeight := FliHdr.height; if FHeight = 0 then FHeight := 200;

    if FliHdr.typ_ = $AF11 then
    begin
      fSpeed := (FliHdr.speed and $FFFF) * 1000 div 70;
    end
      else
    begin
      fSpeed := FliHdr.speed;
      if FliHdr.oframe1 > Stream.Position then
        Stream.Position := FliHdr.oframe1;
    end;
    //--------------------------------------------------------------------------
    // start loading chunks
    Bmp := NewDIBBitmap( FliHdr.width, FliHdr.height, pf8bit );
    TRY

      while Stream.Position < DWORD( Min( Stream.Size, StartPos + FliHdr.size ) ) do
      begin
        if Stream.Read( Chunk, Sizeof( Chunk ) ) <> Sizeof( Chunk ) then Exit;
        NxPos := Stream.Position - Sizeof( Chunk ) + Chunk.sz;
        if NxPos < Stream.Position then NxPos := Stream.Position;
        CASE Chunk.hd OF
        $F1FA: // frame chunk
          begin
            Stream.Read( nsubchunks, 2 );
            Stream.Read( FrameDelay, 2 );
            Stream.Position := Stream.Position + 6;
            while (nsubchunks > 0) and
                  (Stream.Position < DWORD( Min( Stream.Size, NxPos ) )) do
            begin
              if Stream.Read( Chunk, Sizeof( Chunk ) ) <> Sizeof( Chunk ) then Exit;
              Nx1 := Stream.Position - Sizeof( Chunk ) + Chunk.sz;
              if Nx1 < Stream.Position then Nx1 := Stream.Position;
              CASE Chunk.hd OF
              4: FLI_COLOR(0); // 256
              7: FLI_SS2;
              11: FLI_COLOR(2); //64
              12: FLI_LC;
              13: FLI_BLACK;
              15: FLI_BRUN;
              16: FLI_COPY;
              18: FLI_PSTAMP;
              END;
              if LoadPostageStop and Assigned( FPostage ) then Exit;
              Stream.Position := Nx1;
              Dec( nsubchunks );
            end;
            if FFrames = nil then FFrames := NewList;
            Frm := NewBitmap( 0, 0 );
            Frm.Assign( Bmp );
            if FliHdr.typ_ = $AF11 then
              Frm.Tag := FrameDelay * 1000 div 70
            else
              Frm.Tag := FrameDelay;
            FFrames.Add( Frm );
            if LoadPostageStop then Exit;
            Stream.Position := NxPos;
          end;
        else
          Stream.Position := NxPos;
        END;
      end;
    FINALLY
      Bmp.Free;
      Result := (FFrames <> nil) and (FFrames.Count > 0) or
                LoadPostageStop and (FPostage <> nil);
    END;
  FINALLY
    if not Result then
      Stream.Position := StartPos;
  END;
end;

procedure TFlic.StretchDraw(DC: HDC; const Dest: TRect);
var Bmp: PBitmap;
begin
  if (Frame >= 0) and (Frame < Count) then
  begin
    Bmp := FFrames.Items[ Frame ];
    Bmp.StretchDraw( DC, Dest );
  end;
end;

procedure TFlic.StretchDrawTransp(DC: HDC; const Dest: TRect);
begin
  StretchDraw( DC, Dest );
end;

end.
